from typing import Iterable, Optional, Set, Tuple, Union

from spacy.tokens import Doc, Span

from spikex.matcher import Matcher

from ..util import span_idx2i


class AbbrX:
    """
    *Strongly based on scispacy's AbbreviationDetector*.
    Detect abbreviations which are acronyms or by using the algorithm in
    "A simple algorithm for identifying abbreviation definitions in biomedical
    text.", (Schwartz & Hearst, 2003).

    This class sets the `._.abbrs` attribute on spaCy Doc.

    The abbreviations attribute is a `List[Span]` where each Span has the `Span._.long_form`
    attribute set to the long form definition of the abbreviation.

    Note that this class does not replace the spans, or merge them.
    """

    def __init__(self, vocab) -> None:
        Doc.set_extension("abbrs", default=[], force=True)
        Span.set_extension("long_form", default=None, force=True)

        self._matcher = Matcher(vocab)
        self._matcher.add(
            "abbrs",
            [
                # Pattern for abbreviations not enclosed in brackets
                # here we limit to alpha chars only as it could
                # get many exceptions
                [{"IS_ALPHA": True, "IS_UPPER": True, "LENGTH": {">": 1}}],
                # Pattern for abbreviations enclosed in brackets
                # here we try to allow non alpha chars too as it is
                # the more likely standard way to introduce an abbreviation
                [
                    {"TEXT": {"IN": ["(", "["]}},
                    {"OP": "+"},
                    {"TEXT": {"IN": [")", "]"]}},
                ],
            ],
        )

    def find(self, span: Span, doc: Doc) -> Tuple[Span, Set[Span]]:
        """
        Functional version of calling the matcher for a single span.
        This method is helpful if you already have an abbreviation which
        you want to find a definition for.
        """
        dummy_matches = [(-1, int(span.start), int(span.end))]
        filtered = _filter_matches(dummy_matches, doc)
        abbrs = list(self.find_matches_for(filtered, doc))

        if not abbrs:
            return span, set()
        return abbrs[0]

    def __call__(self, doc: Doc) -> Doc:
        matches = self._matcher(doc)
        matches_no_punct = set(
            [
                (
                    x[0],
                    x[1] + (1 if doc[x[1]].is_punct else 0),
                    x[2] - (1 if doc[x[2] - 1].is_punct else 0),
                )
                for x in matches
            ]
        )
        filtered = _filter_matches(matches_no_punct, doc)
        occurences = _find_matches_for(filtered, doc)

        for long_form, short_form in occurences:
            short_form._.long_form = long_form
            doc._.abbrs.append(short_form)
        return doc


def find_abbreviation(
    long_form_candidate: Span, short_form_candidate: Span
) -> Tuple[Span, Optional[Span]]:
    """
    Implements an abbreviation detection algorithm which merges an
    acronym resolution algorithm with detection algorithm based on
    "A simple algorithm for identifying abbreviation definitions in
    biomedical text.", (Schwartz & Hearst, 2003).

    Parameters
    ----------
    long_form_candidate: Span, required.
        The spaCy span for the long form candidate of the definition.
    short_form_candidate: Span, required.
        The spaCy span for the abbreviation candidate.

    Returns
    -------
    A Tuple[Span, Union[Span, None]], representing the short form
    abbreviation and the span corresponding to the long form expansion,
    or None if a match was not found.
    """
    long_form = "".join([x.text_with_ws for x in long_form_candidate])
    short_form = "".join([x.text_with_ws for x in short_form_candidate])
    long_index = len(long_form) - 1
    short_index = len(short_form) - 1
    bounds_idx = _find_abbreviation(
        long_form=long_form,
        long_index=long_index,
        short_form=short_form,
        short_index=short_index,
    )
    if not bounds_idx:
        return
    start_idx, end_idx = bounds_idx
    start, end = span_idx2i(long_form_candidate, start_idx, end_idx)
    return (
        long_form_candidate[start:end],
        short_form_candidate,
    )


def _find_abbreviation(
    *, long_form: str, long_index: int, short_form: str, short_index: int
) -> Union[Tuple[int, int], None]:
    # An abbreviation char must match the starting
    # char of a word (acronym) or an its internal one.
    # In the latter case, the next abbreviation char
    # must be another internal char of the same word
    # or its beginning. In any case, a word which have
    # matched an internal char must have been matched
    # also at its starting char. Moreover, the first char
    # of the abbreviation must be the starting char of a word.
    short_index_reset = short_index
    long_index_end = long_index  # alnum bounds
    has_internal_match = False
    jumps = {}
    while short_index >= 0 and long_index >= 0:
        # Get next abbreviation char to check
        short_char = short_form[short_index].lower()
        # Don't check non alphabetic characters
        if not short_char.isalpha():
            short_index -= 1
            continue
        long_char = long_form[long_index].lower()
        # Don't let there be many unabbreviated words
        if long_char.isspace():
            jump = jumps.setdefault(short_index, 0)
            jump += 1
            if jump > 2:
                break
            jumps[short_index] = jump
        is_starting_char = (
            True
            if long_index == 0
            else not long_form[long_index - 1].isalnum()
        )
        if long_char != short_char:
            # Shrink bounds as the long form
            # ends with non-alphanumeric chars
            if long_index == long_index_end and not long_char.isalnum():
                long_index_end -= 1
            # A word which have matched an internal char
            # must match also its starting char
            if is_starting_char and has_internal_match:
                short_index = short_index_reset
                has_internal_match = False
                continue
            long_index -= 1
            continue
        # First abbreviation char must match
        # the starting char of a word
        if short_index == 0 and not is_starting_char:
            long_index -= 1
            continue
        long_index -= 1
        short_index -= 1
        if not is_starting_char:
            has_internal_match = True
    # In case it didn't end at the starting
    # of a word, move it a step ahead
    if long_index >= 0 and not long_form[long_index].isalnum():
        long_index += 1
    # In case the abbreviation doesn't fully match
    # or it doesn't match from a starting char
    # finding fails
    if (
        short_index >= 0
        or long_index > 0
        and long_form[long_index - 1].isalnum()
    ):
        return
    long_start = max(long_index, 0)
    long_end = long_index_end + 1
    if long_start == long_end:
        return
    return long_start, long_end


def _filter_matches(
    matcher_output: Iterable[Tuple[int, int, int]], doc: Doc
) -> Iterable[Tuple[Span, Span]]:
    # Filter into two cases:
    # 1. <Short Form> (<Long Form>)
    # 2. <Long Form> (<Short Form>) [this case is most common].
    candidates = []
    for match in matcher_output:
        start = match[1]
        end = match[2]
        # Ignore spans with more than 8 words in.
        if end - start > 8:
            continue
        if end - start > 1:
            # Long form is inside the parens.
            # Take two words before.
            short_form_candidate = doc[start - 2 : start - 1]
            if _short_form_filter(short_form_candidate):
                candidates.append((doc[start:end], short_form_candidate))
        else:
            # Normal case.
            # Short form is inside the parens.
            # Sum character lengths of contents of parens.
            abbreviation_length = sum([len(x) for x in doc[start:end]])
            max_words = min(abbreviation_length + 5, abbreviation_length * 2)
            long_start = max(start - max_words - 1, 0)
            long_end = start
            # Skip whether there is no span
            if long_end <= long_start:
                continue
            # Look up to max_words backwards
            long_form_candidate = doc[long_start:long_end]
            candidates.append((long_form_candidate, doc[start:end]))
    return candidates


def _find_matches_for(
    filtered: Iterable[Tuple[Span, Span]], doc: Doc
) -> Iterable[Tuple[Span, Set[Span]]]:
    matches = []
    seen = {}
    start_seen = {}
    global_matcher = Matcher(doc.vocab)
    for (long_candidate, short_candidate) in filtered:
        abbreviation = find_abbreviation(long_candidate, short_candidate)
        # We look for abbreviations, so...
        if abbreviation is None:
            continue
        global_search = {}
        long_form, short_form = abbreviation
        if short_form.start not in start_seen:
            global_search[short_form] = long_form
        if long_form.start not in start_seen:
            global_search[long_form] = short_form
        # Look for each new abbreviation globally to find lone ones
        for form_seen, form_search in global_search.items():
            start_seen.setdefault(form_seen.start)
            # We've already seen this
            if form_seen.text in seen:
                continue
            seen.setdefault(form_seen.text, form_seen)
            pattern = [{"TEXT": t.text} for t in form_search]
            global_matcher.add(form_seen.text, [pattern])
        matches.append((long_form, short_form))
    # Search for lone abbreviations globally
    for key, start, end in global_matcher(doc):
        # We've already found this match
        if start in start_seen:
            continue
        text = doc.vocab.strings[key]
        # Never seen, skip
        if text not in seen:
            continue
        already_seen = seen[text]
        global_match = Span(doc, start, end)
        start_seen.setdefault(global_match.start)
        seen.setdefault(global_match.text, global_match)
        # Short form should be the shortest
        if len(already_seen) < len(global_match):
            short_form = already_seen
            long_form = global_match
        else:
            short_form = global_match
            long_form = already_seen
        matches.append((long_form, short_form))
    yield from sorted(matches, key=lambda x: x[0].start)


def _short_form_filter(span: Span) -> bool:
    # All words are between length 2 and 10
    if not all([2 <= len(x) < 10 for x in span]):
        return False
    # At least one word is alpha numeric
    if not any([x.is_alpha for x in span]):
        return False
    return True



import gzip
import io
from math import ceil, floor
from pathlib import Path
from typing import List, Union

import srsly
from spacy.tokens import Doc, Span, Token


def span_idx2i(
    source: Union[Doc, Span, List[Token]],
    start_idx: int,
    end_idx: int,
    maxlen: int = None,
):
    start_i = 0 if start_idx == 0 else -1
    maxlen = maxlen or (
        sum([len(t) for t in source])
        if isinstance(source, list)
        else len(source.text)
    )
    end_i = len(source) if end_idx == maxlen else -1
    if start_i >= 0 and end_i >= 0:
        return start_i, end_i
    start_i = idx2i(source, start_idx)
    if end_i >= 0:
        return start_i, end_i
    last_i = start_i
    offset_i = source[0].i
    offset_idx = source[0].idx
    for token in source[start_i:]:
        last_i = token.i - offset_i
        if token.idx - offset_idx < end_idx:
            continue
        return start_i, last_i
    token = source[last_i]
    last_idx = token.idx - offset_idx
    if end_idx < last_idx:
        return start_i, last_i
    if end_idx < last_idx + len(token):
        return start_i, last_i + 1
    return start_i, start_i + 1


def idx2i(source: Union[Doc, Span, List[Token]], idx: int):
    max_i = len(source)
    offset_idx = source[0].idx

    def _idx2i(_idx, start, end):
        if start == end:
            return start
        dhalf = (start + end) / 2
        fhalf = floor(dhalf)
        if fhalf >= max_i:
            return fhalf - 1
        if fhalf == 0:
            return 0
        chalf = ceil(dhalf)
        if chalf >= max_i:
            return chalf - 1
        fidx = source[fhalf].idx - offset_idx
        cidx = source[chalf].idx - offset_idx
        if _idx <= fidx + 1:
            coord = (fhalf, fidx)
        elif _idx >= cidx - 1:
            coord = (chalf, cidx)
        elif _idx <= fidx + len(source[fhalf]):
            _idx = fidx
            coord = (fhalf, fidx)
        elif _idx <= cidx + len(source[chalf]):
            _idx = cidx
            coord = (chalf, cidx)
        else:
            return
        if coord[1] == _idx or coord[1] - 1 == _idx:
            return coord[0]
        if coord[1] + 1 == _idx:
            return coord[0] + 1
        if coord[1] > _idx:
            return _idx2i(_idx, start, coord[0])
        else:
            return _idx2i(_idx, coord[0], end)

    return _idx2i(idx, 0, max_i)


def pickle_dumps(data, protocol=None):
    return srsly.pickle_dumps(data, protocol=protocol)


def pickle_dump(data, path, protocol=None, compress=None):
    open = gzip.open if compress else io.open
    with open(path, "wb") as fd:
        fd.write(pickle_dumps(data, protocol=protocol))


def pickle_loads(data):
    return srsly.pickle_loads(data)


def pickle_load(path: Path):
    with path.open("rb") as fd:
        # The first two bytes of a gzip file are: 1f 8b
        is_gz = fd.read(2) == b"\x1f\x8b"
    open = gzip.open if is_gz else io.open
    with open(path, "rb") as fd:
        return pickle_loads(fd.read())



import regex as re


class Fragment:
    """
    A fragment of text that ends with a possible sentence boundary
    """

    def __init__(self, tokens):
        self.tokens = tokens
        self.next = None
        self.label = None
        self.features = None
        self.prediction = None
        self.is_sent_end = False

    def __str__(self):
        s = "".join([t.text_with_ws for t in self.tokens])
        if self.is_sent_end:
            s += " <EOS> "
        return s

    @property
    def first_token(self):
        if not self.tokens:
            return
        return self.tokens[0]

    @property
    def last_token(self):
        if not self.tokens:
            return
        return self.tokens[-1]

    def words(self, clean=None):
        return [
            _clean(token.text) if clean is not None else token.text
            for token in self.tokens
        ]


def _clean(text):
    # normalize numbers, discard some punctuation that can be ambiguous
    t = re.sub(r"[.,\d]*\d", "<NUM>", text)
    t = re.sub(r"[^a-zA-Z0-9,.;:<>\-'\/?!$% ]", "", t)
    t = t.replace("--", " ")  # sometimes starts a sentence... trouble
    return t



from spacy.tokens import Doc, Span

from ..matcher import Matcher


class PhraseX:
    """
    Create a custom `Doc`'s underscore extension in which stores
    matches found by applying given pattern matching expressions.
    """

    def __init__(self, vocab, phrases_name, patterns):
        Doc.set_extension(phrases_name, default=[], force=True)
        self._phrases_name = phrases_name
        self._matcher = Matcher(vocab)
        self._matcher.add(phrases_name, patterns)

    def __call__(self, doc: Doc):
        phrases = []
        last_start = 0
        for _, start, end in self._matcher(doc):
            if last_start >= end:
                continue
            last_start = end
            span = Span(doc, start, end)
            phrases.append(span)
        phrases.sort(key=lambda x: (x.start, -len(x)))
        setattr(doc._, self._phrases_name, _fix_overlappings(phrases))
        return doc


NP_PATTERNS = [
    [
        {
            "POS": {"IN": ["ADJ", "ADV", "DET", "NUM", "PROPN"]},
            "OP": "*",
        },
        {"POS": {"IN": ["ADP", "CONJ", "CCONJ"]}, "OP": "?"},
        {
            "POS": {
                "IN": ["ADJ", "ADP", "ADV", "NOUN", "NUM", "PRON", "PROPN"]
            },
            "OP": "*",
        },
        {
            "POS": {"IN": ["NOUN", "PROPN"]},
        },
    ]
]


class NounPhraseX(PhraseX):
    """
    Detect noun phrases and assigns them at the custom underscore attribute `noun_phrases`
    """

    def __init__(self, vocab):
        super(NounPhraseX, self).__init__(vocab, "noun_phrases", NP_PATTERNS)


VP_PATTERNS = [
    [
        {"POS": {"IN": ["ADV", "AUX", "PART", "VERB"]}, "OP": "*"},
        {"POS": {"IN": ["AUX", "VERB"]}},
    ]
]


class VerbPhraseX(PhraseX):
    """
    Detect verb phrases and assigns them at the custom underscore attribute `verb_phrases`
    """

    def __init__(self, vocab):
        super(VerbPhraseX, self).__init__(vocab, "verb_phrases", VP_PATTERNS)


def _fix_overlappings(spans):
    good_spans = set()
    for span in spans:
        should_add_span = False
        for other_span in spans:
            # good if spans are identical
            # or they aren't overlapping
            if (
                span.start == other_span.start
                and span.end == other_span.end
                or span.start >= other_span.end
                or span.end <= other_span.start
            ):
                should_add_span = True
                continue
            # exit loop as spans overlap
            # but one is larger
            if (
                span.start > other_span.start
                and span.end <= other_span.end
                or span.start >= other_span.start
                and span.end < other_span.end
            ):
                should_add_span = False
                break
        if should_add_span:
            good_spans.add(span)
    return sorted(good_spans, key=lambda x: (x.start, -len(x)))



import math
from pathlib import Path
from typing import Dict, List

import regex as re
from wasabi import msg

from spikex.util import pickle_dump, pickle_load

from .fragment import Fragment


class NBModel:
    """
    Naive Bayes model, with a few tweaks:
    - all feature types are pooled together for normalization (this might help
      because the independence assumption is so broken for our features)
    - smoothing: add 0.1 to all counts
    - priors are modified for better performance (this is mysterious but works much better)
    """

    _PRIOR_FEAT = "<prior>"

    def __init__(self):
        self.feats = {}
        self.lower_words = {}
        self.non_abbrs = {}

    @staticmethod
    def load(path: Path = None):
        if path is None:
            path = Path(__file__).parent / "nbmodel.gz"
        data = pickle_load(path)
        model = NBModel()
        model.feats = data["feats"]
        model.lower_words = data["lower_words"]
        model.non_abbrs = data["non_abbrs"]
        return model

    def save(self, path: Path):
        data = {
            "feats": self.feats,
            "lower_words": self.lower_words,
            "non_abbrs": self.non_abbrs,
        }
        pickle_dump(data, path / "nbmodel.gz", compress=True)

    def featurize_one(self, frag: Fragment):
        frag.features = self._get_features(frag)
        return frag.features

    def featurize(self, corpus: List[Fragment]):
        for frag in corpus:
            self.featurize_one(frag)

    def classify_one(self, frag: Fragment):
        # the prior is weird, but it works better this way, consistently
        probs = {
            label: self.feats[label, self._PRIOR_FEAT] ** 4 for label in [0, 1]
        }
        for label in probs:
            if frag.features is None:
                self.featurize_one(frag)
            for feat, val in frag.features.items():
                key = (label, feat + "_" + val)
                if key not in self.feats:
                    continue
                probs[label] *= self.feats[key]
        frag.prediction = _normalize(probs)[1]
        return frag.prediction

    def classify(self, fragments: List[Fragment]):
        for frag in fragments:
            self.classify_one(frag)

    def train(self, corpus: List[Fragment], verbose: bool = None):
        if not corpus:
            raise ValueError
        msg.no_print = not verbose
        with msg.loading("setting things up..."):
            self._setup_training(corpus)
        msg.text("train Naive Bayes model")
        feats = {}
        totals = {}
        for frag in corpus:
            for feat, val in frag.features.items():
                feats[frag.label][feat + "_" + val] += 1
            totals[frag.label] += len(frag.features)
        # add-1 smoothing and normalization
        with msg.loading("smoothing... "):
            smooth_inc = 0.1
            all_feat_names = set(feats[True].keys()).union(
                set(feats[False].keys())
            )
            for label in [0, 1]:
                totals[label] += len(all_feat_names) * smooth_inc
                for feat in all_feat_names:
                    feats[label][feat] += smooth_inc
                    feats[label][feat] /= totals[label]
                    self.feats[(label, feat)] = feats[label][feat]
                feats[label][self._PRIOR_FEAT] = (
                    totals[label] / totals.totalCount()
                )
                self.feats[(label, self._PRIOR_FEAT)] = feats[label][
                    self._PRIOR_FEAT
                ]
        msg.good("done")

    def _setup_training(self, corpus: List[Fragment]):
        self.lower_words, self.non_abbrs = corpus.get_stats(verbose=False)
        self.lower_words = dict(self.lower_words)
        self.non_abbrs = dict(self.non_abbrs)
        self.featurize(corpus)

    def _get_features(self, frag: Fragment):
        # ... w1. (sb?) w2 ...
        # Features, listed roughly in order of importance:
        #   (1) w1: word that includes a period
        #   (2) w2: the next word, if it exists
        #   (3) w1length: number of alphabetic characters in w1
        #   (4) w2cap: true if w2 is capitalized
        #   (5) both: w1 and w2
        #   (6) w1abbr: log count of w1 in training without a final period
        #   (7) w2lower: log count of w2 in training as lowercased
        #   (8) w1w2upper: w1 and w2 is capitalized
        words1 = frag.words(clean=True)
        w1 = words1[-1] if words1 else ""
        words2 = frag.next.words(clean=True) if frag.next else []
        w2 = words2[0] if words2 else ""
        c1 = re.sub(r"(^.+?\-)", "", w1)
        c2 = re.sub(r"(\-.+?)$", "", w2)
        feats = {}
        feats["w1"] = c1
        feats["w2"] = c2
        feats["both"] = c1 + "_" + c2
        len1 = min(10, len(re.sub(r"\W", "", c1)))
        if c1.replace(".", "").isalpha():
            feats["w1length"] = str(len1)
            try:
                feats["w1abbr"] = str(
                    int(math.log(1 + self.non_abbrs[c1[:-1]]))
                )
            except:
                feats["w1abbr"] = str(int(math.log(1)))
        if c2.replace(".", "").isalpha():
            feats["w2cap"] = str(c2[0].isupper())
            try:
                feats["w2lower"] = str(
                    int(math.log(1 + self.lower_words[c2.lower()]))
                )
            except:
                feats["w2lower"] = str(int(math.log(1)))
            feats["w1w2upper"] = c1 + "_" + str(c2[0].isupper())
        return feats


def _normalize(counter: Dict[str, float]):
    # normalize a counter by dividing each value by the sum of all values
    total = float(sum(counter.values()))
    return {k: v / total for k, v in counter.items()}



from typing import Union

import regex as re
from spacy.attrs import DEP, LEMMA, POS, TAG, intify_attr
from spacy.errors import Errors, MatchPatternError
from spacy.tokens import Doc, Span, Token

from ..defaults import spacy_version

if spacy_version < 3:
    from spacy.util import get_json_validator, validate_json

    from ._schemas import TOKEN_PATTERN_SCHEMA
else:
    # Pattern validation changed as of spaCy 3.0
    from spacy.attrs import MORPH  # type: ignore
    from spacy.schemas import validate_token_pattern  # type: ignore

from functools import partial

from ._schemas import TOKEN_PATTERN_SCHEMA

__all__ = ["Matcher"]

DocLike = Union[Doc, Span]


class Matcher(object):
    def __init__(self, vocab, validate=False):

        self._specs = {}
        self._patterns = {}
        self._callbacks = {}
        self._seen_attrs = set()
        self.vocab = vocab
        self.validate = validate

        try:
            self._validator = partial(
                validate_json,
                validator=get_json_validator(TOKEN_PATTERN_SCHEMA),
            )
        except NameError:
            self._validator = validate_token_pattern

    def __len__(self):
        """
        Get the number of rules added to the matcher.
        Note that this only returns the number of rules (identical with the number of IDs),
        not the number of individual patterns.

        Returns
        ------
        int
            The number of rules.
        """
        return len(self._patterns)

    def __contains__(self, key: str):
        """
        Check whether the matcher contains rules for a key.

        Parameters
        ----------
        key: str
            The key.

        Returns
        -------
        bool
            Whether the matcher contains rules for this key.
        """
        key = self._normalize_key(key)
        return key in self._patterns

    def __getitem__(self, key: str):
        """
        Retrieve the pattern stored for a key.
        A KeyError is raised if the key does not exist.

        Parameters
        ----------
        key: str
            The key to retrieve.

        Returns
        -------
        tuple
            The rule, as an (on_match, patterns) tuple.
        """
        key = self._normalize_key(key)
        return (self._callbacks[key], self._patterns[key])

    def get(self, key: str, default=None):
        """
        Retrieve the pattern stored for a key,
        or default if it does not exist.

        Parameters
        ----------
        key: str
            The key to retrieve.

        Returns
        -------
        tuple:
            The rule, as an (on_match, patterns) tuple.
        """
        key = self._normalize_key(key)
        return self[key] if key in self else (default, default)

    def add(self, key, patterns, on_match=None):
        """
        Add a match-rule to the matcher.
        A match-rule consists of:
            an ID key,
            a list of patterns,
            an on_match callback.
        If the key exists, the patterns are appended to the
        previous ones, and the previous on_match callback is replaced. The
        `on_match` callback will receive the arguments `(matcher, doc, i, matches)`.
        You can also set `on_match` to `None` to not perform any action.
        A pattern consists of one or more `token_specs`, where a `token_spec`
        is a dictionary mapping attribute IDs to values, and optionally a
        quantifier operator under the key "op". The available quantifiers are:
        '!': Negate the pattern, by requiring it to match exactly 0 times.
        '?': Make the pattern optional, by allowing it to match 0 or 1 times.
        '+': Require the pattern to match 1 or more times.
        '*': Allow the pattern to zero or more times.
        The + and * operators are usually interpretted "greedily", i.e. longer
        matches are returned where possible. However, if you specify two '+'
        and '*' patterns in a row and their matches overlap, the first
        operator will behave non-greedily. This quirk in the semantics makes
        the matcher more efficient, by avoiding the need for back-tracking.

        Parameters
        ----------
        key: str
            The match ID.
        patterns: list
            The patterns to add for the given key.
        on_match: callable
            Optional callback executed on match.
        """
        errors = {}
        if on_match is not None and not hasattr(on_match, "__call__"):
            raise ValueError(Errors.E171.format(arg_type=type(on_match)))
        for i, pattern in enumerate(patterns):
            if len(pattern) == 0:
                raise ValueError(Errors.E012.format(key=key))
            if not isinstance(pattern, list):
                raise ValueError(Errors.E178.format(pat=pattern, key=key))
            if self.validate:
                errors[i] = self._validator(pattern)
        if any(err for err in errors.values()):
            raise MatchPatternError(key, errors)
        key = self._normalize_key(key)
        self._specs.setdefault(key, [])
        for i, pattern in enumerate(patterns):
            try:
                patternspec = _preprocess_pattern(pattern)
            except ValueError as err:
                raise MatchPatternError(key, {i: [str(err)]})
            self._specs[key].append(patternspec)
            for token in pattern:
                for attr in token:
                    iattr = intify_attr(attr)
                    self._seen_attrs.add(iattr)
        self._patterns.setdefault(key, [])
        self._callbacks[key] = on_match
        self._patterns[key].extend(patterns)

    def remove(self, key: str):
        """
        Remove a rule from the matcher.
        A ValueError is raised if the key does not exist.

        Parameters
        ----------
        key: str
            The ID of the match rule.
        """
        key = self._normalize_key(key)
        if key not in self._patterns:
            raise ValueError(Errors.E175.format(key=key))
        self._specs.pop(key)
        self._patterns.pop(key)
        self._callbacks.pop(key)

    def __call__(self, doclike: DocLike, allow_missing=False):
        """
        Find all token sequences matching the supplied patterns.

        Parameters
        ----------
        doc: Doc
            The document to match over.

        Returns
        -------
        list
            A list of `(match_id, start, end)` tuples,
            describing the matches. A match tuple describes a span
            `doc[start:end]`.
        """
        if not allow_missing:
            # legacy between spaCy versions
            attr2pipe = {
                TAG: ("tagger", "is_tagged"),
                POS: ("morphologizer", "is_tagged"),
                LEMMA: ("lemmatizer", "is_tagged"),
                DEP: ("parser", "is_parsed"),
            }
            if spacy_version >= 3:
                attr2pipe[MORPH] = ("morphologizer", None)
            for attr, (pipe, flag) in attr2pipe.items():
                if (
                    attr not in self._seen_attrs
                    or (spacy_version >= 3 and doclike.has_annotation(attr))
                    or (spacy_version < 3 and getattr(doclike, flag))
                ):
                    continue
                raise ValueError(
                    Errors.E155.format(
                        pipe=pipe, attr=self.vocab.strings.as_string(attr)
                    )
                )
        matches = []
        seen = set()
        for match in _find_matches(doclike, self._specs):
            if match in seen:
                continue
            seen.add(match)
            matches.append(match)
        for i, match in enumerate(matches):
            on_match = self._callbacks.get(match[0], None)
            if on_match is not None:
                on_match(self, doclike, i, matches)
        return matches

    def _normalize_key(self, key):
        if isinstance(key, int):
            return key
        return (
            self.vocab.strings.add(key)
            if key not in self.vocab.strings
            else self.vocab.strings[key]
        )


def _find_matches(tokens, specs):
    attrs_maps_cache = {}
    num_tokens = len(tokens)
    for key, pattern_specs in specs.items():
        for pattern_spec, anchor_gs in pattern_specs:
            candidates = [((0, num_tokens), {})]
            for attr, (xp, is_ext) in pattern_spec.items():
                if attr not in attrs_maps_cache:
                    attrs_maps_cache[attr] = _attr_maps(attr, tokens, is_ext)
                i2idx, idx2i, text = attrs_maps_cache[attr]
                maxlen = len(text)
                new_candidates = []
                for candidate, anchor_ss in candidates:
                    start_idx = i2idx[candidate[0]]
                    end_idx = i2idx[candidate[1]]
                    curr_text = text[start_idx:end_idx]
                    for match in xp.finditer(curr_text, overlapped=True):
                        span = (
                            start_idx + match.span()[0],
                            start_idx + match.span()[1],
                        )
                        start, end = _span_idx2i(span, idx2i, maxlen)
                        new_ss = {}
                        for i in range(len(match.groups())):
                            group_i = i + 1
                            if group_i not in anchor_gs:
                                continue
                            span_g = match.span(group_i)
                            span = (
                                start_idx + span_g[0],
                                start_idx + span_g[1],
                            )
                            new_ss[group_i] = _span_idx2i(span, idx2i, maxlen)
                        if anchor_ss:
                            should_stop = False
                            for group_i, span in new_ss.items():
                                if anchor_ss[group_i] != span:
                                    should_stop = True
                                    break
                            if should_stop:
                                continue
                        new_candidates.append(((start, end), new_ss))
                candidates = new_candidates
            matches = [c[0] for c in candidates]
            yield from (
                (key, *match) for match in _filter_out_submatches(matches)
            )


def _attr_maps(attr, tokens, is_extension):
    i2idx = {}
    idx2i = {}
    text_tokens = []
    curr_length = 0
    num_spaces = 0
    regex_attr = attr == "REGEX"
    for i, token in enumerate(tokens):
        pad = i if not regex_attr else num_spaces
        idx = curr_length + pad
        i2idx[i] = idx
        idx2i[idx] = i
        if is_extension:
            value = token._.get(attr)
        else:
            value = _get_token_attr(token, attr)
        value = str(value)
        curr_length += len(value)
        if regex_attr:
            value += token.whitespace_
            num_spaces += len(token.whitespace_)
        else:
            num_spaces = i
        text_tokens.append(value)
    curr_length += num_spaces
    i2idx[len(tokens)] = curr_length
    idx2i[curr_length] = len(tokens)
    text = ("" if regex_attr else " ").join(text_tokens)
    return (i2idx, idx2i, text)


def _span_idx2i(span_idx, idx2i, maxlen):
    start = span_idx[0]
    while start not in idx2i and start < maxlen:
        start += 1
    end = span_idx[1]
    while end not in idx2i and end < maxlen:
        end += 1
    return idx2i[start], idx2i[end]


def _filter_out_submatches(matches):
    i1 = 0
    num_matches = len(matches)
    while i1 < num_matches:
        yield matches[i1]
        i2 = i1 + 1
        while (
            i2 < num_matches
            and matches[i1][0] < matches[i2][0]
            and matches[i1][1] == matches[i2][1]
        ):
            i2 += 1
        i1 = i2


def _preprocess_pattern(pattern):
    pattern_spec = {}
    num_tokens = len(pattern)
    for tokens_spec in pattern:
        if not isinstance(tokens_spec, dict):
            raise ValueError(Errors.E154.format())
        for attr, value in {**tokens_spec}.items():
            if not (
                isinstance(value, str)
                or isinstance(value, bool)
                or isinstance(value, int)
                or isinstance(value, dict)
            ):
                raise ValueError(
                    Errors.E153.format(vtype=type(value).__name__)
                )
            # normalize attributes
            if attr.islower():
                del tokens_spec[attr]
                attr = attr.upper()
                tokens_spec[attr] = value
            if attr == "SENT_START":
                del tokens_spec[attr]
                attr = "IS_SENT_START"
                tokens_spec[attr] = value
            elif attr == "OP":
                continue
            is_extension = attr == "_"
            if is_extension and not isinstance(value, dict):
                raise ValueError(Errors.E154.format())
            if not is_extension and isinstance(value, dict):
                for k, v in {**value}.items():
                    if k.isalpha() and k.isupper():
                        continue
                    del value[k]
                    value[k.upper()] = v
            for a in value.keys() if is_extension else [attr]:
                pattern_spec.setdefault(a, ([None] * num_tokens, is_extension))
    for i, tokens_spec in enumerate(pattern):
        _align_tokens_spec(pattern_spec, tokens_spec, i)
    return _finalize_pattern_spec(pattern_spec)


# Quantifiers
_NONE = "x"
_ONE = "1"
_ONE_PLUS = "+"
_ZERO = "!"
_ZERO_ONE = "?"
_ZERO_PLUS = "*"

# Regex
_XP_ONE_TOKEN = r"[^\s]+"
_XP_TOKEN_START = r"(?:\s|^)"
_XP_TOKEN_DELIM = r"(?:\s|^|$)"

# Predicates
_REGEX_PREDICATES = ("REGEX",)
_SETMEMBER_PREDICATES = ("IN", "NOT_IN")
_COMPARISON_PREDICATES = ("==", "!=", ">=", "<=", ">", "<")

# Other
_ANCHOR_QS = (_ONE, _ONE_PLUS)


def _finalize_pattern_spec(spec):
    anchor_gs = set()
    final_spec = {}
    for attr, (xps, is_extension) in spec.items():
        if not anchor_gs:
            for i, (_, q) in enumerate(xps):
                if q not in _ANCHOR_QS:
                    continue
                anchor_gs.add(i + 1)
        if attr in _REGEX_PREDICATES:
            regex = "".join([x[0] for x in xps])
        else:
            regex = "".join([_XP_TOKEN_START, *(x[0] for x in xps)])
        flags = re.U | re.M
        if attr in ("LENGTH", "LOWER"):
            flags |= re.I
        final_spec[attr] = (re.compile(regex, flags=flags), is_extension)
    sort_by = lambda x: x[0] not in ("LEMMA", "LOWER", "TEXT")
    final_spec = {k: v for k, v in sorted(final_spec.items(), key=sort_by)}
    return (final_spec, anchor_gs)


def _align_tokens_spec(spec, tokens_spec, index):
    xp_cond_delim = f"(?({index + 1}){_XP_TOKEN_DELIM}|)"
    for a1, (xp, q) in _attrs_spec_from_tokens_spec(tokens_spec):
        fix_q = q if q != _ZERO else _ONE
        needs_delim = fix_q == _ONE
        for a2, (xps, _) in spec.items():
            _q = _xp = None
            if not a1 or a1 == a2:
                _q = q
                _xp = xp
            elif xps[index] == None:
                _q = fix_q
                _xp = _XP_ONE_TOKEN
            if not (_q and _xp):
                continue
            lazy = _q == _ONE_PLUS and _xp == _XP_ONE_TOKEN
            _xp = _re_wrap_quantifier(_q, _xp, lazy)
            if needs_delim:
                _xp += xp_cond_delim
            xps[index] = (_xp, _q)


def _attrs_spec_from_tokens_spec(tokens_spec):
    if not tokens_spec:
        yield None, (_XP_ONE_TOKEN, _ONE)
    elif "REGEX" in tokens_spec:
        yield "REGEX", (tokens_spec["REGEX"], _NONE)
    else:
        q = tokens_spec.get("OP")
        if q and len(tokens_spec) == 1:
            yield None, (_XP_ONE_TOKEN, q)
        else:
            if not q:
                q = _ONE  # default quantifier
            for (attr, xp) in _attrs_xp_from_tokens_spec(tokens_spec):
                yield attr, (xp, q)


def _attrs_xp_from_tokens_spec(tokens_spec, skip_validate=False):
    for attr, value in tokens_spec.items():
        if attr == "OP":
            continue
        if attr == "_":
            yield from _attrs_xp_from_tokens_spec(value, True)
            continue
        if (
            not skip_validate
            and attr not in TOKEN_PATTERN_SCHEMA["items"]["properties"]
        ):
            raise ValueError(Errors.E152.format(attr=attr))
        if isinstance(value, int):
            value = str(value)
        if isinstance(value, dict):
            for p, a in value.items():
                if p in _REGEX_PREDICATES:
                    yield attr, _xp_from_regex(a)
                elif p in _SETMEMBER_PREDICATES:
                    yield attr, _xp_from_setmember(p, a)
                elif p in _COMPARISON_PREDICATES:
                    yield attr, _xp_from_comparison(p, a)
                else:
                    raise ValueError()
            continue
        yield attr, (
            re.escape(value) if attr not in _REGEX_PREDICATES else value
        )


def _xp_from_regex(regex):
    # Symbols `^` and `$` for start, end string must be removed
    # Here we don't use `re.sub` because works bad with raw strings
    split = re.split(r"(?:(?<!\[|\\)\^|(?<!\\)\$)", regex)
    merge = "".join(split)
    return merge if len(split) > 1 else "".join([r"[^ ]*?", merge, r"[^ ]*?"])


def _xp_from_setmember(operator, args):
    # We optimize in case of a unique argument
    if len(args) == 1:
        pipe = re.escape(args[0])
    else:
        terms = (re.escape(term) for term in args)
        pipe = "".join([r"(?:", r"|".join(terms), r")"])
    return f"(?!{pipe})[^ ]+" if operator == "NOT_IN" else pipe


def _xp_from_comparison(operator, length):
    return _re_wrap_length(operator, length)


_WRAP_Q_LOOKUP = {
    _NONE: "({xp})",
    _ONE: "({xp})",
    _ONE_PLUS: "((?:{{xp}}{delim})+)".format(delim=_XP_TOKEN_DELIM),
    _ZERO: "(?!{xp})([^ ]+)",
    _ZERO_ONE: "({{xp}}{delim})?".format(delim=_XP_TOKEN_DELIM),
    _ZERO_PLUS: "((?:{{xp}}{delim})*)".format(delim=_XP_TOKEN_DELIM),
}


def _re_wrap_quantifier(q, xp, lazy=False):
    if q is None:
        return xp
    if lazy and q not in (_ONE_PLUS, _ZERO_PLUS):
        raise ValueError
    if q not in _WRAP_Q_LOOKUP:
        keys = ", ".join(_WRAP_Q_LOOKUP.keys())
        raise ValueError(Errors.E011.format(op=q, opts=keys))
    xpq = _WRAP_Q_LOOKUP[q].format(xp=xp)
    if not lazy:
        return xpq
    return "".join([xpq[:-1], "?", xpq[-1:]])


def _re_wrap_length(cmp, l):
    if cmp == "==":
        return "(?:[^ ]{{{}}}+)".format(l)
    elif cmp == "!=":
        return "(?:[^ ]{{1,{}}}+|[^ ]{{{},}}+)".format(l - 1, l + 1)
    elif cmp == ">=":
        return "(?:[^ ]{{{},}}+)".format(l)
    elif cmp == "<=":
        return "(?:[^ ]{{1,{}}}+)".format(l)
    elif cmp == ">":
        return "(?:[^ ]{{{},}}+)".format(l + 1)
    elif cmp == "<":
        return "(?:[^ ]{{1,{}}}+)".format(l - 1)
    else:
        raise ValueError(
            Errors.E126.format(bad=cmp, good=_COMPARISON_PREDICATES)
        )


def _get_token_attr(token: Token, attr: str):
    if attr == "REGEX":
        return token.text
    elif attr == "LEMMA":
        return token.lemma_.lower()
    elif attr == "NORM":
        if not token.norm_:
            return token.lex.norm
        return token.norm_
    elif attr == "POS":
        return token.pos_
    elif attr == "TAG":
        return token.tag_
    elif attr == "DEP":
        return token.dep_
    elif attr == "SENT_START":
        return token.sent_start
    elif attr == "ENT_TYPE":
        return token.ent_type_
    elif attr == "ORTH":
        return token.orth_
    elif attr == "TEXT":
        return token.text
    elif attr == "LOWER":
        return token.lower_
    elif attr == "SHAPE":
        return token.shape_
    elif attr == "PREFIX":
        return token.prefix_
    elif attr == "SUFFIX":
        return token.suffix_
    # LENGTH attribute must be checked on text
    # and it cannot live together with
    # another textual attribute, so we set it
    # as LOWER for a performance reason
    elif attr == "LENGTH":
        return token.lower_
    elif attr == "CLUSTER":
        return token.cluster
    elif attr == "LANG":
        return token.lang_
    elif token.check_flag(intify_attr(attr)):
        return "True"
    return "False"



from spacy.tokens import Doc, Span

from ..wikigraph import WikiGraph


class WikiPageX:
    """
    Detect spans which match with Wikipedia pages, based on a
    specific `WikiGraph`.
    """

    def __init__(self, wikigraph: WikiGraph) -> None:
        Doc.set_extension("wiki_spans", default=[])
        Span.set_extension("wiki_pages", default=[])
        self._wg = wikigraph

    def __call__(self, doc: Doc):
        idx2i, text = _preprocess_doc(doc)
        maxtlen = len(text)
        for start_idx, end_idx, pages in self._wg.find_pages(text):
            start_i, end_i = _span_idx2i(start_idx, end_idx, idx2i, maxtlen)
            if start_i >= end_i:
                continue
            span = doc[start_i:end_i]
            for p in pages:
                vx = self._wg.get_vertex(p)
                span._.wiki_pages.append((p, vx["title"]))
            doc._.wiki_spans.append(span)
        return doc


def _preprocess_doc(doc):
    idx2i = {}
    curr_idx = 0
    text_tokens = []
    for i, token in enumerate(doc):
        idx2i[curr_idx] = i
        value = token.lemma_ if token.pos_ == "NOUN" else token.text
        value += token.whitespace_
        text_tokens.append(value)
        curr_idx += len(value)
    idx2i[curr_idx] = len(doc)
    text = "".join(text_tokens)
    return idx2i, text.replace(" ", "_")


def _span_idx2i(start_idx, end_idx, idx2i, maxlen):
    while start_idx not in idx2i and start_idx < maxlen:
        start_idx += 1
    while end_idx not in idx2i and end_idx < maxlen:
        end_idx += 1
    if start_idx == end_idx:
        end_i = idx2i[end_idx]
        return end_i - 1, end_i
    return idx2i[start_idx], idx2i[end_idx]



from os import path, walk
from pathlib import Path
from shutil import copy

from setuptools import setup
from srsly import json_loads


def list_files(data_dir):
    output = []
    for root, _, filenames in walk(data_dir):
        for filename in filenames:
            if not filename.startswith("."):
                output.append(path.join(root, filename))
    output = [path.relpath(p, path.dirname(data_dir)) for p in output]
    output.append("meta.json")
    return output


def setup_package():
    root_path = Path(__file__).parent.absolute()
    meta_path = root_path / "meta.json"
    meta = json_loads(meta_path.read_text())
    graph_name = meta["name"]
    pkg_data = list_files(path.join(graph_name, meta["fullname"]))
    requirements = [f"spikex{meta['spikex_version']}"]
    copy(meta_path, path.join(graph_name))
    setup(
        name=graph_name,
        description=meta["description"],
        author=meta["author"],
        author_email=meta["email"],
        url=meta["url"],
        version=meta["version"],
        license=meta["license"],
        packages=[graph_name],
        package_data={graph_name: pkg_data},
        install_requires=requirements,
        zip_safe=False,
    )


if __name__ == "__main__":
    setup_package()



from collections import Counter
from itertools import combinations
from random import randrange
from typing import List

import numpy as np
from gensim.models import KeyedVectors
from spacy.tokens import Doc, Span


class ClusterX:
    """
    Cluster `noun_chunks` of a `Doc` by applying a revisited **Radial Ball Mapper** algorithm.
    """

    def __init__(self, min_score: float):
        Doc.set_extension("cluster_chunks", default=[])
        self.min_score = min_score

    def __call__(self, doc: Doc):
        doc._.cluster_chunks = cluster_chunks(
            list(doc.noun_chunks), min_score=self.min_score
        )
        return doc


def cluster_chunks(
    chunks: List[Span],
    stopwords: bool = False,
    filter_pos: List[str] = None,
    min_score: float = None,
):
    """
    Cluster chunks by using a revisited **Radial Ball Mapper** algorithm

    Parameters
    ----------
    chunks : List[Span]
        Chunks to cluster.
    stopwords : bool, optional
        Flag to exclude stopwords from chunks, by default False.
    filter_pos : List[str], optional
        POS tags to filter chunk words, by default None
    min_score : float, optional
        Threshold for clustering chunks, by default None

    Returns
    -------
    List[List[Span]]
        Clusters of chunks
    """
    key2index, key2vector = _map_key_to_vector(chunks, stopwords, filter_pos)
    if not key2index or not key2vector:
        return
    model = KeyedVectors(chunks[0].vector.size)
    keys = list(key2vector.keys())
    weights = list(key2vector.values())
    model.add(keys, weights)
    clusters = cluster_balls_multi(model, keys, min_score=min_score)
    return [[chunks[key2index[i]] for i in cluster] for cluster in clusters]


def cluster_balls(
    model: KeyedVectors,
    root: str = None,
    max_size: int = None,
    min_score: float = None,
):
    """
    Cluster a model's keys by applying a revisited Radial Ball Mapper algorithm.

    A root key should be specified in case a point of interest is known.
    Not specifying any root key, a random one is picked from the model.

    If no otherwise specified, a `max_size` of 30 is used by default.

    if no otherwise specified, a `min_score` calculated as mean of all best similarities,
    minus a gap of 0.05, is used by default.

    Parameters
    ----------
    model : KeyedVectors
        Word2Vec model which stores all keys and vectors.
    root : str
        Point of interest from which to start clustering balls, by default None.
    max_size : int, optional
        Maximum size of a ball in terms of number of keys, by default None.
    min_score : float, optional
        Minimum similarity threshold for starting a cluster, by default None.

    Returns
    -------
    List[List[str]]
        Clusters of keys
    """
    if root is None:
        rand_i = randrange(0, len(model.index2entity))
        root = model.index2entity[rand_i]
    elif root not in model:
        return
    max_size = max_size or 30
    neighs = model.most_similar(root, topn=max_size)
    if not neighs:
        return
    if min_score is None:
        mean = _get_neighs_mean_score(model, neighs)
        min_score = mean - 0.05
    cut_off = 0.5
    clusters = []
    root_cluster = {root}
    seen = {root: (root_cluster, 1)}
    for n, s in neighs:
        if s <= cut_off:
            break
        if n in seen:
            continue
        if s >= min_score:
            root_cluster.add(n)
            seen.setdefault(n, (root_cluster, s))
            continue
        cluster = set()
        min_sub_score = min_score + 0.05
        for nn, ss in model.most_similar(n, topn=max_size):
            if ss <= cut_off:
                break
            if nn in seen:
                c, b = seen[nn]
                if c == root_cluster or b >= ss:
                    continue
            if ss >= min_sub_score:
                if nn in seen:
                    prev_cluster = seen[nn][0]
                    prev_cluster.remove(nn)
                cluster.add(nn)
                seen[nn] = (cluster, ss)
        cluster.add(n)
        seen.setdefault(n, (cluster, 1))
        clusters.append(cluster)
        if len(cluster) < 3:
            continue
        intruder = _get_intruder(model, cluster)
        if intruder is None:
            continue
        del seen[intruder]
        cluster.remove(intruder)
    clusters.insert(0, root_cluster)
    return clusters


def cluster_balls_multi(
    model: KeyedVectors,
    keys: List[str],
    max_size: int = None,
    min_score: float = None,
):
    """
    Cluster a model's keys by applying a revisited Radial Ball Mapper algorithm
    to each key of a list, fixing overlappings in order to build coherent clusters.

    This provides a method to create clusters based on multiple points of interest instead of one only.

    Parameters
    ----------
    model : KeyedVectors
        Word2Vec model which stores all keys and vectors.
    keys : List[str]
        Keys to use as points of interest.
    max_size : int, optional
        Maximum size of a ball in terms of number of keys, by default None.
    min_score : float, optional
        Minimum similarity threshold for starting a cluster, by default None.

    Returns
    -------
    List[List[str]]
        Clusters of keys
    """
    clusters = []
    for key in keys:
        for ball in cluster_balls(
            model, key, max_size=max_size, min_score=min_score
        ):
            i = 0
            merged = False
            to_remove = set()
            for i in range(len(clusters)):
                cluster = clusters[i]
                if (
                    ball == cluster
                    or len(set.intersection(ball, cluster)) == 0
                ):
                    continue
                if ball.issuperset(cluster):
                    to_remove.add(i)
                    continue
                merged = True
                if ball.issubset(cluster):
                    continue
                merge = set.union(ball, cluster)
                if merge in clusters:
                    continue
                clusters[i] = merge
            if not merged and ball not in clusters:
                clusters.append(ball)
            clusters = [
                el for i, el in enumerate(clusters) if i not in to_remove
            ]
    return clusters


def _map_key_to_vector(chunks, stopwords=None, filter_pos=None):
    key2index = {}
    key2vector = {}
    for index, chunk in enumerate(chunks):
        vectors = []
        key = chunk.text.lower()
        for token in chunk:
            if (
                stopwords
                and token.is_stopword
                or filter_pos
                and token.pos_ not in filter_pos
            ):
                continue
            if not np.any(token.vector) or token.vector.size == 0:
                continue
            vectors.append(token.vector)
        if not vectors:
            continue
        key2index.setdefault(key, index)
        key2vector.setdefault(key, sum(vectors) / len(vectors))
    return key2index, key2vector


def _get_neighs_mean_score(model, neighs):
    scores = []
    for neigh, _ in neighs:
        similar = model.most_similar(neigh, topn=1)
        if not similar:
            continue
        _, score = similar[0]
        scores.append(score)
    return sum(scores) / len(scores)


def _get_intruder(model, cluster):
    intruders = Counter()
    maxlen = len(cluster) - 1
    for c in combinations(cluster, maxlen):
        intruder = model.doesnt_match(c)
        intruders.update([intruder])
        if intruders[intruder] == maxlen:
            return intruder



import regex as re
from spacy.tokens import Doc

from .fragment import Fragment
from .nbmodel import NBModel


class SentX:
    def __init__(self):
        self.model = NBModel.load()

    def __call__(self, doc: Doc):
        fragments = _get_fragments(doc)
        self.model.classify(fragments)
        start_token_ids = _sent_start_token_ids(fragments)
        for token in doc:
            token.is_sent_start = token.i in start_token_ids
        return doc


def _sent_start_token_ids(fragments):
    # iterate all tokens that end a sentence
    # according to predictions or labels
    thresh = 0.5
    is_next_start = True
    start_token_ids = []
    for frag in fragments:
        if is_next_start:
            start_token_ids.append(frag.first_token.i)
            # continue
        is_next_start = (
            frag.label > thresh or frag.prediction > thresh or frag.is_sent_end
        )
    return start_token_ids


SAFE_ACRONYMS = re.compile(
    r"^(tm|TM|Std|Nor|Fig|FIG|Figs|FIGS|pat|Pat|ser|Ser|No|Num|eg|[1-9][0-9])$"
)
SAFE_ABBRS = re.compile(r"(?:^[a-zA-Z]{1,3}\.)")
SPECIAL_SENT_STARTERS = re.compile(
    r"((?:[a-zèéòàìù]+|[A-ZÈÉÒÙÀÌ]{3,}|[0-9]\s*\))(?:\s*\))?)?\s*"
    r"((?:The|This|That|Those|These|Who|When|What|Which|Where|Whose)\s*"
    r"(?:[a-z]|[A-Z][a-zèéòàìù]+\s+[a-z])?)"
)


def _get_fragments(doc):
    fragments = []
    curr_tokens = []
    curr_frag = None
    last_frag = None
    for token in doc:
        curr_tokens.append(token)
        if len(curr_tokens) == 1:
            continue
        prev_token = curr_tokens[-2]
        split_at_last = SPECIAL_SENT_STARTERS.search(token.text) and (
            (
                prev_token.pos_
                and prev_token.pos_ in ("NOUN", "PROPN", "ADJ", "ADV")
                or prev_token.tag_
                and prev_token.tag_ in ("NN", "NNS", "NNP", "NNPS", "JJ", "RB")
            )
            or "\n" in prev_token.text
        )
        if not _is_sentence_boundary(token, prev_token) and not split_at_last:
            continue
        last_frag = curr_frag
        if split_at_last:
            curr_tokens = curr_tokens[:-1]
        curr_frag = Fragment([*curr_tokens])
        curr_frag.is_sent_end = True
        curr_frag.label = int("<S>" in token.text)
        curr_tokens = []
        if split_at_last:
            curr_tokens.append(token)
        if last_frag is not None:
            last_frag.next_frag = curr_frag
            fragments.append(last_frag)
    # last frag
    if len(curr_tokens) > 0:
        if curr_frag is not None:
            fragments.append(curr_frag)
        curr_frag = Fragment(curr_tokens)
        curr_frag.label = int("<S>" in curr_tokens[-1].text)
    if curr_frag is not None:
        curr_frag.is_sent_end = True
        fragments.append(curr_frag)
    return fragments


def _is_sentence_boundary(token, prev_token):
    c = _unannotate(token.text)
    prev_word = prev_token.text
    if SAFE_ACRONYMS.search(prev_word) or SAFE_ABBRS.search(prev_word):
        return
    if (c.endswith(".") or re.match(r".*\.[\"')\]]*$", c)) and not (
        SAFE_ACRONYMS.search(c) or SAFE_ABBRS.search(c)
    ):
        return True


def _unannotate(t):
    # get rid of a tokenized word's annotations
    return re.sub(r"(<A>)?(<E>)?(<S>)?$", "", t)



import pstats
from cProfile import Profile
from pathlib import Path

import memory_profiler as mp
from spacy import load as spacy_load
from srsly import read_jsonl

from ..matcher import Matcher
from ..wikigraph import load as wg_load


def profile_matcher(text_path: Path, patterns_path: Path, memory: bool = None):
    def func():
        nlp = spacy_load("en_core_web_sm")
        doc = nlp(text_path.read_text())
        matcher = Matcher(doc.vocab)
        patterns = [
            p["pattern"] if "pattern" in p else p
            for p in read_jsonl(patterns_path)
        ]
        matcher.add("Profile", patterns)
        matcher(doc)

    _profile(func, memory)


def profile_wikigraph_load(graph_name: str, memory: bool = None):
    def func():
        _ = wg_load(graph_name)

    _profile(func, memory)


def profile_wikigraph_exec(graph_name: str, memory: bool = None):
    wg = wg_load(graph_name)
    text = """
    Hong Kong (CNN) China's top military commander in Hong Kong has emphasized the role of the People's Liberation Army (PLA) in upholding "national sovereignty" in the city a day ahead of expected anti-government protests.

    Chen Daoxiang, the PLA commander in Hong Kong, was speaking days after China announced plans to introduce a draconian new national security law which threatens many of the semi-autonomous city's civil liberties and political freedoms.
    "Garrison officers and soldiers are determined, confident, and capable of safeguarding national sovereignty and development interests and the long-term prosperity and stability of Hong Kong," Chen said Tuesday, adding that the Hong Kong Garrison viewed the national security law as "conducive to deter separatist forces and external intervention."
    Almost every official department in Hong Kong has issued statements of support for the proposed law this week -- from the police to the fire department and the Government Flying Service. The PLA Garrison has also made similar statements about its role in the past.
    Chen's words do carry some extra weight, however, given the uneasy status of the PLA in the city.
    """

    def func():
        for _ in wg.find_pages(text):
            pass

    _profile(func, memory)


def _profile(fn, memory=None):
    if memory:
        _mem_profile(fn)
    else:
        _time_profile(fn)


def _mem_profile(fn):
    prof = mp.LineProfiler(backend="psutil")
    prof(fn)()
    mp.show_results(prof)


def _time_profile(fn):
    profiler = Profile()
    profiler.runcall(fn)
    stats = pstats.Stats(profiler)
    stats.sort_stats("time")
    stats.print_stats(40)



import zlib
from contextlib import closing
from functools import partial
from multiprocessing import Pool
from os import cpu_count
from pathlib import Path

import regex as re
import requests
from ftfy import fix_text
from smart_open.compression import compression_wrapper
from smart_open.http import open as http_open
from tqdm import tqdm
from wasabi import msg
from yarl import URL

from spikex.util import pickle_dumps, pickle_loads

__all__ = [
    "iter_page_dump_data",
    "iter_redirect_dump_data",
    "iter_pagelinks_dump_data",
    "iter_categorylinks_dump_data",
]

config = {
    "dumps_path": None,
    "max_workers": cpu_count(),
    "verbose": None,
    "version": "latest",
    "wiki": "en",
}

WIKI_NS_KIND_PAGE = "0"
WIKI_NS_KIND_CATEGORY = "14"

WIKI_CL_TYPE_PAGE = "page"
WIKI_CL_TYPE_CATEGORY = "subcat"

_WIKI_DP_PAGE = "page"
_WIKI_DP_REDIRECT = "redirect"
_WIKI_DP_PAGELINKS = "pagelinks"
_WIKI_DP_PAGE_PROPS = "page_props"
_WIKI_DP_CATEGORYLINKS = "categorylinks"

_WIKI_DL_NAME = "{w}wiki-{v}-{t}.sql.gz"
_WIKI_BASE_DL_PATH = "https://dumps.wikimedia.org/{w}wiki/"
_WIKI_DL_PATH = _WIKI_BASE_DL_PATH + "{v}/{n}"


def resolve_version(wiki, version):
    url = _WIKI_BASE_DL_PATH.format(w=wiki)
    response = requests.get(url)
    versions = re.findall(r"href=\"(\d+)/\"", response.text)
    if version in versions:
        return version
    if version == "latest":
        return max(versions)
    msg.fail(
        "Wikipedia dump version not found",
        f"Pick one of these: {', '.join(versions)}.",
        exits=1,
    )


def iter_page_props_dump_data(**kwargs):
    dump_url = _get_wiki_dump_dl_url(_WIKI_DP_PAGE_PROPS, **kwargs)
    return _parse_wiki_sql_dump(dump_url, _parse_fx_page_props_dump, **kwargs)


def _parse_fx_page_props_dump(el):
    pageid = el[0]
    prop = el[1]
    value = el[2]
    return pageid, prop, value


def iter_page_dump_data(**kwargs):
    dump_url = _get_wiki_dump_dl_url(_WIKI_DP_PAGE, **kwargs)
    return _parse_wiki_sql_dump(dump_url, _parse_fx_page_dump, **kwargs)


def _parse_fx_page_dump(el):
    if el[1] not in (WIKI_NS_KIND_CATEGORY, WIKI_NS_KIND_PAGE):
        return
    ns_kind = el[1]
    pageid = el[0]
    title = _unescape_title(fix_text(el[2]))
    return ns_kind, pageid, title


def iter_redirect_dump_data(**kwargs):
    dump_url = _get_wiki_dump_dl_url(_WIKI_DP_REDIRECT, **kwargs)
    yield from _parse_wiki_sql_dump(
        dump_url, _parse_fx_redirect_dump, **kwargs
    )


def _parse_fx_redirect_dump(el):
    if el[1] not in (WIKI_NS_KIND_CATEGORY, WIKI_NS_KIND_PAGE):
        return
    redirectid = el[0]
    title = _unescape_title(fix_text(el[2]))
    return redirectid, title


def iter_pagelinks_dump_data(**kwargs):
    dump_url = _get_wiki_dump_dl_url(_WIKI_DP_PAGELINKS, **kwargs)
    return _parse_wiki_sql_dump(dump_url, _parse_fx_pagelinks_dump, **kwargs)


def _parse_fx_pagelinks_dump(el):
    if el[1] != WIKI_NS_KIND_PAGE or el[3] != WIKI_NS_KIND_PAGE:
        return
    sourceid = el[0]
    target = _unescape_title(fix_text(el[2]))
    return sourceid, target


def iter_categorylinks_dump_data(**kwargs):
    dump_url = _get_wiki_dump_dl_url(_WIKI_DP_CATEGORYLINKS, **kwargs)
    return _parse_wiki_sql_dump(
        dump_url, _parse_fx_categorylinks_dump, **kwargs
    )


def _parse_fx_categorylinks_dump(el):
    if el[6] not in (WIKI_CL_TYPE_CATEGORY, WIKI_CL_TYPE_PAGE):
        return
    cl_type = el[6]
    sourceid = el[0]
    target = _unescape_title(fix_text(el[1]))
    return cl_type, sourceid, target


def _unescape_title(title):
    return title.replace("\\'", "'").replace('\\"', '"')


def _get_wiki_dump_dl_url(t, **kwargs):
    _kwargs = {**config, **kwargs}
    w = _kwargs["wiki"]
    v = _kwargs["version"]
    n = _WIKI_DL_NAME.format(w=w, t=t, v=v)
    dp = _kwargs["dumps_path"]
    if dp is not None and dp.exists():
        path = dp.joinpath(n)
        if path.exists():
            return path
    return URL(_WIKI_DL_PATH.format(w=w, n=n, v=v))


def _get_wiki_dump_obj(wiki_sql_dump_url, verbose=None):
    if isinstance(wiki_sql_dump_url, Path):
        if not wiki_sql_dump_url.exists():
            raise FileNotFoundError
        compress_obj = wiki_sql_dump_url.open("rb")
        content_len = wiki_sql_dump_url.stat().st_size
        if content_len == 0:
            raise FileNotFoundError
    elif isinstance(wiki_sql_dump_url, URL):
        compress_obj = http_open(str(wiki_sql_dump_url), mode="rb")
        content_len = int(compress_obj.response.headers.get("content-length"))
    else:
        raise ValueError
    msg.text(f"from: {wiki_sql_dump_url}", show=verbose)
    return compress_obj, content_len


def _parse_wiki_sql_dump(wiki_sql_dump_url, parse_fx, **kwargs):
    _kwargs = {**config, **kwargs}
    dumps_path = _kwargs["dumps_path"]
    max_workers = _kwargs["max_workers"]
    verbose = _kwargs["verbose"]
    compress_bytes_read = 0
    dump_name = wiki_sql_dump_url.name
    msg.text(f"-> {dump_name}", show=verbose)
    tqdm_disable = not verbose
    tqdm_kwargs = {
        "unit": "B",
        "unit_scale": True,
        "unit_divisor": 1024,
        "disable": tqdm_disable,
    }
    compress_obj, content_len = _get_wiki_dump_obj(wiki_sql_dump_url, verbose)
    should_reopen_compress_obj = False
    if dumps_path is not None:
        if not dumps_path.exists():
            dumps_path.mkdir()
        dump_filepath = dumps_path.joinpath(dump_name)
        if not dump_filepath.exists() or dump_filepath.stat().st_size == 0:
            with tqdm(
                desc="download to disk",
                total=content_len,
                **tqdm_kwargs,
            ) as pbar, dump_filepath.open("wb") as fd:
                bytes_read = 0
                for chunk in compress_obj:
                    fd.write(chunk)
                    compress_bytes = compress_obj.tell()
                    pbar.update(compress_bytes - bytes_read)
                    bytes_read = compress_bytes
            compress_obj.close()
            should_reopen_compress_obj = True
            wiki_sql_dump_url = dump_filepath
    if should_reopen_compress_obj:
        compress_obj, content_len = _get_wiki_dump_obj(wiki_sql_dump_url)
    with tqdm(
        desc="parse",
        total=content_len,
        **tqdm_kwargs,
    ) as pbar, compression_wrapper(compress_obj, "rb") as decompress_obj:
        compress_bytes_read = 0
        with closing(Pool(max_workers)) as pool:
            task = partial(_parsing_task, parse_fx=parse_fx)
            for res in pool.imap_unordered(task, decompress_obj, chunksize=10):
                compress_bytes = compress_obj.tell()
                pbar.update(compress_bytes - compress_bytes_read)
                compress_bytes_read = compress_bytes
                yield from pickle_loads(zlib.decompress(res))
    msg.good(dump_name, show=verbose)


def _parsing_task(line, parse_fx):
    ret = []
    for el in _parse_wiki_sql_dump_line(line):
        res = parse_fx(el)
        if res is None:
            continue
        ret.append(res)
    return zlib.compress(pickle_dumps(ret, protocol=-1))


def _parse_wiki_sql_dump_line(line):
    line = line.decode("latin1")
    if line.startswith("INSERT INTO"):
        el_end = 0
        el_start = 0
        curr_tuple = []
        is_escaping = False
        is_string_open = False
        line_start = line.index("(")
        for i in range(line_start, len(line)):
            c = line[i]
            if not is_string_open:
                if c == "(":
                    el_start = i + 1
                    continue
                if c == ")":
                    el_end = i if el_end == 0 else el_end
                    el = "" if el_start == el_end else line[el_start:el_end]
                    curr_tuple.append(el)
                    el_end = 0
                    el_start = 0
                    if curr_tuple:
                        yield tuple(curr_tuple)
                        curr_tuple = []
                    continue
                if c == ",":
                    if el_start > 0:
                        el_end = i if el_end == 0 else el_end
                        el = (
                            "" if el_start == el_end else line[el_start:el_end]
                        )
                        curr_tuple.append(el)
                        el_end = 0
                        el_start = i + 1
                    continue
                if c == "'":
                    el_start = i + 1
                    is_string_open = True
                    continue
            else:
                if c == "\\":
                    is_escaping = not is_escaping
                elif c == "'" and not is_escaping:
                    el_end = i
                    is_string_open = False
                    continue
                else:
                    is_escaping = False



from pathlib import Path

from srsly import json_loads

_here_path = Path(__file__).parent
pkg_path = _here_path.joinpath("pkg")


def get_meta():
    meta_path = _here_path.joinpath("meta.json")
    return json_loads(meta_path.read_bytes())



from .create import create_wikigraph
from .download import download_wikigraph
from .package import package_wikigraph
from .profile import (
    profile_matcher,
    profile_wikigraph_exec,
    profile_wikigraph_load,
)



import shutil
from distutils.dir_util import copy_tree
from os import rename
from pathlib import Path

from srsly import json_loads
from wasabi import msg

from ..templates.wikigraph import pkg_path


def package_wikigraph(input_path: Path, output_path: Path, force: bool = None):
    """
    Generate an installable Python package for a `WikiGraph`.

    After packaging, "python setup.py sdist" must be run in the package directory,
    which will create a .tar.gz archive that can be installed via "pip install".

    Parameters
    ----------
    input_path : Path
        [description]
    output_path : Path
        [description]
    force : bool, optional
        [description], by default None
    """
    if not input_path or not input_path.exists():
        msg.fail("Can't locate graph data", input_path, exits=1)
    if not output_path:
        msg.fail("Output directory is missing", output_path, exits=1)
    if not output_path.exists():
        output_path.mkdir()
        msg.good("Created output directory: {}".format(output_path))
    meta_path = input_path / "meta.json"
    if not meta_path.exists():
        msg.fail("Can't find graph meta.json", meta_path, exits=1)
    meta = json_loads(meta_path.read_text())
    graph_fullname = meta["fullname"]
    package_path = output_path / graph_fullname
    if package_path.exists():
        if not force:
            msg.fail(
                "Package directory already exists",
                "Please delete the directory and try again, or use the "
                "`--force` flag to overwrite existing "
                "directories.".format(path=package_path),
                exits=1,
            )
        shutil.rmtree(package_path)
    package_path.mkdir()
    shutil.copy(meta_path, package_path)
    copy_tree(str(pkg_path), str(package_path))
    graph_name = meta["name"]
    rename(package_path / "graph-name", package_path / graph_name)
    module_path = package_path / graph_name
    copy_tree(str(input_path), str(module_path / graph_fullname))
    msg.good(
        "Successfully created package {}".format(graph_name), package_path
    )
    msg.text(
        "To build the package, run `python setup.py sdist` in this directory."
    )



from pathlib import Path

from srsly import json_dumps
from wasabi import msg

from .. import __version__ as spikex_version
from ..templates.wikigraph import get_meta
from ..wikigraph import WikiGraph


def create_wikigraph(
    output_path: Path,
    wiki="en",
    version="latest",
    dumps_path: Path = None,
    max_workers: int = None,
    silent: bool = None,
    force: bool = None,
):
    """
    Create a `WikiGraph` from a specific dump.

    It can then be used by directly loading it, or
    it can be packaged with the `package-wikigraph` command.

    Parameters
    ----------
    output_path : Path
        Path in which to store the `WikiGraph`.
    wiki : str, optional
        Wikipedia dump type to use, by default "en".
    version : str, optional
        Wikipedia dump version to use, by default "latest".
    dumps_path : Path, optional
        Path in which to find previously downloaded dumps,
        or where to save dumps downloaded in this call, by default None.
    max_workers : int, optional
        Maximum number of processes to use, by default None.
    silent : bool, optional
        Do not print anything in stout, by default None.
    force : bool, optional
        Overwrite content in output_path, if any, by default None.
    """
    if not output_path.exists():
        output_path.mkdir()
        msg.good(f"Created output directory: {output_path}")
    graph_name = f"{wiki}wiki_core"
    graph_path = output_path.joinpath(graph_name)
    if not force and graph_path.exists():
        msg.fail(
            f"Output path already contains {graph_name} directory",
            "Use --force to overwrite it",
            exits=1,
        )
    kwargs = {
        "dumps_path": dumps_path,
        "max_workers": max_workers,
        "wiki": wiki,
        "version": version,
        "verbose": not silent,
    }
    wg = WikiGraph.build(**kwargs)
    if not graph_path.exists():
        graph_path.mkdir()
    graph_format = "picklez"
    with msg.loading("dump to disk..."):
        wg.dump(graph_path, graph_format=graph_format)
    meta = get_meta()
    meta["name"] = graph_name
    meta["version"] = wg.version
    meta["graph_format"] = graph_format
    meta["spikex_version"] = f">={spikex_version}"
    meta["fullname"] = f"{graph_name}-{spikex_version}"
    meta["sources"].append("Wikipedia")
    meta_path = graph_path.joinpath("meta.json")
    meta_path.write_text(json_dumps(meta, indent=2))
    msg.good(f"Successfully created {graph_name}.")



from spacy.tokens import Doc, Span, Token

from ..matcher import Matcher


class LabelX:
    """
    Detect labelings, which are pattern matching expressions
    linked to a specific label, in case using abbreviations and
    acronyms previously found by AbbrX pipe.

    If `only_longest` is flagged, it fixes overlapping matches
    by merging them and assigning the label of the last match.
    """

    def __init__(
        self, vocab, labelings=None, validate=None, only_longest=None
    ):
        Doc.set_extension("labelings", default=[], force=True)
        Token.set_extension("labels", default=[], force=True)
        self._matcher = Matcher(vocab, validate)
        self._only_longest = only_longest
        if not labelings or labelings is None:
            return
        for label, patterns in labelings:
            self.add(label, patterns)

    def add(self, label, patterns, on_match=None):
        """
        Add a labeling rule to the labeler.

        Parameters
        ----------
        label: str
            The label related to patterns.
        patterns: list
            The pattern matching expressions.
        on_match: callable, optional
            Callback executed on match, by default None.
        """
        self._matcher.add(label, patterns, on_match=on_match)

    def __call__(self, doc: Doc):
        """
        Collect all labels related to patterns matching tokens.
        Each token have a `labels` extension in which to store them.
        The supplied `Doc` have a `labelings` extension in which
        all labeled spans are collected.
        If the doc has abbrs, they contribute to label spans.

        Parameters
        ----------
        doc: Doc
            The doc to label over.

        Returns
        -------
        Doc
            The doc after labeling.
        """
        for key, start, end in self._matcher(doc):
            label = doc.vocab.strings[key]
            span = Span(doc, start, end, label)
            for token in span:
                if label in token._.labels:
                    continue
                token._.labels.append(label)
            doc._.labelings.append(span)
        _sort_labelings(doc)
        if doc.has_extension("abbrs"):
            _merge_abbrs_labelings(doc)
        if self._only_longest:
            _fix_overlabelings(doc)
        return doc


def _merge_abbrs_labelings(doc):
    num_labelings = len(doc._.labelings)
    chunk2label = {s.text: s.label for s in doc._.labelings}
    for abbr in doc._.abbrs:
        # first check if long form is labeled
        # and short one is not
        if (
            abbr._.long_form.text in chunk2label
            and abbr.text not in chunk2label
        ):
            label = chunk2label[abbr._.long_form.text]
            short_span = Span(doc, abbr.start, abbr.end, label)
            doc._.labelings.append(short_span)
        # otherwise check if short form is labeled
        # and long one is not
        elif (
            abbr.text in chunk2label
            and abbr._.long_form.text not in chunk2label
        ):
            lf = abbr._.long_form
            label = chunk2label[abbr.text]
            long_span = Span(doc, lf.start, lf.end, label)
            doc._.labelings.append(long_span)
    # sort labelings if there's changes only
    if num_labelings < len(doc._.labelings):
        _sort_labelings(doc)


def _fix_overlabelings(doc):
    good_labelings = set()
    for span in doc._.labelings:
        should_add_span = False
        for other_span in doc._.labelings:
            # good if spans are identical
            # or they aren't overlapping
            if (
                span.start == other_span.start
                and span.end == other_span.end
                or span.start >= other_span.end
                or span.end <= other_span.start
            ):
                should_add_span = True
                continue
            # exit loop as spans overlap
            # but one is larger
            if (
                span.start > other_span.start
                and span.end <= other_span.end
                or span.start >= other_span.start
                and span.end < other_span.end
            ):
                should_add_span = False
                break
            # merge spans overlapping
            # in a tail-head manner
            # (last label wins)
            if (
                span.start < other_span.start
                and span.end > other_span.start
                and span.end < other_span.end
            ) or (
                span.start > other_span.start
                and span.start < other_span.end
                and span.end > other_span.end
            ):
                if span.start < other_span.start:
                    start = span.start
                    end = other_span.end
                    label = other_span.label
                else:
                    start = other_span.start
                    end = span.end
                    label = span.label
                merge_span = Span(doc, start, end, label)
                good_labelings.add(merge_span)
                should_add_span = False
                break
        if should_add_span:
            good_labelings.add(span)
    doc._.labelings = list(good_labelings)


def _sort_labelings(doc):
    doc._.labelings.sort(key=lambda x: (x.start, -len(x)))



from setuptools import setup, find_packages

with open("README.md") as readme_file:
    readme = readme_file.read()

requirements = open("requirements.txt").read().splitlines()

test_requirements = ["pytest"]

setup(
    author="Erre Quadro srl",
    author_email="paolo.arduin@errequadrosrl.com",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="SPIKEX - SpaCy Pipes for Knowledge Extraction",
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="spikex",
    name="spikex",
    packages=find_packages(include=["spikex", "spikex.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/erre-quadro/spikex",
    version="0.4.0",
    zip_safe=False,
    entry_points = {
        'console_scripts': ['spikex=spikex.__main__:main'],
    }
)



import importlib
import json
import logging
from pathlib import Path
from typing import Iterable, Union

import regex as re
from cyac import AC
from igraph import Graph, Vertex
from wasabi import msg

from ..util import pickle_dump, pickle_load
from . import dumptools as dt

logging.basicConfig(
    format="%(asctime)s : %(threadName)s : %(levelname)s : %(message)s",
    level=logging.INFO,
)

VERTEX_KIND_CATEGORY = dt.WIKI_NS_KIND_CATEGORY
VERTEX_KIND_PAGE = dt.WIKI_NS_KIND_PAGE

VertexType = Union[Vertex, int]


def load(name: Union[Path, str]):
    if isinstance(name, Path):
        if not name.exists():
            raise IOError
        data_path = name
        meta_path = name / "meta.json"
        meta = json.load(meta_path.open())
    elif isinstance(name, str):
        kls = importlib.import_module(name)
        data_path = kls.data_path
        meta = kls.meta
    else:
        raise IOError
    f_path = data_path / "f"
    g_path = data_path / "g"
    if not g_path.exists() or not f_path.exists():
        raise IOError
    return WikiGraph.load(g_path, f_path, meta)


class WikiGraph:
    def __init__(self):
        self.g = None
        self.wiki = None
        self.version = None
        self._page_finder = WikiPageFinder()

    @staticmethod
    def build(**kwargs):
        wg = WikiGraph()
        wg.wiki = kwargs["wiki"]
        kwargs["version"] = dt.resolve_version(wg.wiki, kwargs["version"])
        wg.version = kwargs["version"]
        wg.g = make_graph(**kwargs)
        wg.build_page_finder()
        return wg

    @staticmethod
    def load(g_path, f_path, meta):
        wg = WikiGraph()
        wg._page_finder = pickle_load(f_path)
        graph_format = meta["graph_format"]
        wg.g = Graph.Load(g_path.open("rb"), format=graph_format)
        return wg

    def dump(self, dir_path: Path, graph_format: str = None):
        pickle_dump(self._page_finder, dir_path / "f", compress=True)
        self.g.write(dir_path / "g", graph_format or "picklez")

    def build_page_finder(self):
        self._page_finder.build(self.pages())

    def pages(self):
        yield from self.g.vs.select(kind_eq=VERTEX_KIND_PAGE)

    def categories(self):
        yield from self.g.vs.select(kind_eq=VERTEX_KIND_CATEGORY)

    def find_vertex(self, title: str):
        return self.g.vs.find(title=title)

    def get_vertex(self, vertex: VertexType):
        return self.g.vs[vertex] if isinstance(vertex, int) else vertex

    def redirect_vertex(self, vertex: VertexType):
        vx = self.get_vertex(vertex)
        if vx["redirect"] < 0:
            return vx
        return self.g.vs.find(vx["redirect"])

    def get_parent_vertices(self, vertex: VertexType):
        vx = self.redirect_vertex(vertex)
        es = self.g.es.select(_source=vx)
        return [e.target_vertex.index for e in es]

    def get_ancestor_vertices(self, vertex: VertexType, until: int = 2):
        def get_ancestors(vx, left):
            left -= 1
            ancestors = set()
            for parent in self.get_parent_vertices(vx):
                ancestors.add(parent)
                if not left:
                    continue
                ancestors.update(get_ancestors(parent, left))
            return ancestors

        return list(get_ancestors(vertex, until))

    def find_pages(self, text: str):
        yield from self._page_finder.find_pages(text)


_XP_SEPS = re.compile(r"(\p{P})")


class WikiPageFinder:
    def __init__(self, pages: Iterable[Vertex] = None):
        self._ac = None
        self._id2pages = {}
        if pages is not None:
            self.build(pages)

    def __getstate__(self):
        return (self._ac, self._id2pages)

    def __setstate__(self, state):
        self._ac, self._id2pages = state

    def build(self, pages: Iterable[Vertex]):
        key2pages = {}
        for page in pages:
            key = _clean_title(page["title"]).lower()
            key_pages = key2pages.setdefault(key, [])
            key_pages.append(page.index)
        self._ac = AC.build(key2pages.keys(), ignore_case=True)
        for key, id_ in self._ac.items():
            self._id2pages.setdefault(id_, key2pages[key])

    def find_pages(self, text: str):
        def iter_matches(source):
            ac_seps = set([ord(p) for p in _XP_SEPS.findall(source)])
            for id_, start_idx, end_idx in self._ac.match_longest(
                source, ac_seps
            ):
                yield (start_idx, end_idx, self._id2pages[id_])

        for match in iter_matches(text):
            yield match
            end_idx = match[1]
            start_idx = match[0]
            tot_match_tokens = len(_XP_SEPS.split(text[start_idx:end_idx]))
            if tot_match_tokens < 2:
                continue
            submatches = set()
            tokens = _XP_SEPS.split(text[start_idx:])
            chunk = tokens[0]
            for i in range(tot_match_tokens):
                if i > 0:
                    fix_i = i * 2
                    start_idx += len(tokens[fix_i - 1])
                    chunk = "".join(tokens[fix_i:])
                for sidx, eidx, pages in iter_matches(chunk):
                    s = start_idx + sidx
                    if s >= end_idx:
                        break
                    if s in submatches:
                        continue
                    submatches.add(s)
                    yield (s, start_idx + eidx, pages)


def make_graph(**kwargs):
    cat2id = {}
    page2idx = {}
    g = Graph(directed=True)
    pprops = _get_pprops(**kwargs)
    verbose = "verbose" in kwargs and kwargs["verbose"]
    msg_no_print = msg.no_print
    msg.no_print = not verbose
    iter_page_data = dt.iter_page_dump_data(**kwargs)
    for ns_kind, pageid, title in iter_page_data:
        disambi = False
        if pageid in pprops:
            page_props = pprops[pageid]
            if "hiddencat" in page_props or "noindex" in page_props:
                continue
            if "disambiguation" in page_props:
                disambi = True
        vx = g.add_vertex(
            pageid,
            title=title,
            kind=ns_kind,
            redirect=-1,
            disambi=disambi,
        )
        if ns_kind == dt.WIKI_NS_KIND_CATEGORY:
            cat2id[title] = pageid
        elif ns_kind == dt.WIKI_NS_KIND_PAGE:
            page2idx[title] = vx.index
    _add_redirects(g, page2idx, **kwargs)
    _add_category_links(g, page2idx.keys(), cat2id, **kwargs)
    msg.no_print = msg_no_print
    return g


def _get_pprops(**kwargs):
    pprops = {}
    iter_pprops_data = dt.iter_page_props_dump_data(**kwargs)
    for pageid, prop, value in iter_pprops_data:
        page_props = pprops.setdefault(pageid, {})
        page_props.setdefault(prop, value)
    return pprops


def _add_redirects(g, page2idx, **kwargs):
    for source_id, target_title in dt.iter_redirect_dump_data(**kwargs):
        try:
            source_vx = g.vs.find(source_id)
            target_idx = page2idx[target_title]
        except (KeyError, ValueError):
            continue
        source_vx["redirect"] = target_idx


def _add_category_links(g, pages, cat2id, **kwargs):
    edges = []
    for _, source_id, target_title in dt.iter_categorylinks_dump_data(
        **kwargs
    ):
        if target_title not in pages:
            continue
        try:
            source_vx = g.vs.find(source_id)
            if (
                source_vx["kind"] == dt.WIKI_NS_KIND_CATEGORY
                and source_vx["title"] not in pages
            ):
                continue
            target_id = cat2id[target_title]
        except (KeyError, ValueError):
            continue
        edges.append((source_vx["name"], target_id))
    with msg.loading("adding category edges..."):
        g.add_edges(edges)


def _clean_title(title: str):
    open_at = title.find("_(")
    if open_at < 0:
        return title
    close_at = title.find(")", open_at)
    if close_at < 0:
        return title
    a = title[:open_at]
    b = title[close_at + 1 :]
    return "".join((a, b))



import sys

from spacy.util import run_command
from wasabi import msg

WIKIGRAPHS_TABLE = {
    "enwiki_core": "https://errequadrosrl-my.sharepoint.com/:u:/g/personal/paolo_arduin_errequadrosrl_onmicrosoft_com/ESedYiVvufpCtImuOlFXm6MB_5YyfKQnZIvDinnYbL-NmA?Download=1",
    "simplewiki_core": "https://errequadrosrl-my.sharepoint.com/:u:/g/personal/paolo_arduin_errequadrosrl_onmicrosoft_com/EQhheXcD9KtGpXyoZ9a2zOEBmGIvZXuyFoV1KoYOzgsjLw?Download=1",
    "itwiki_core": "https://errequadrosrl-my.sharepoint.com/:u:/g/personal/paolo_arduin_errequadrosrl_onmicrosoft_com/EVBnV0JaBNlFpmNg91hT458BfFjY_7MW2kqIvRCkhdpWVQ?Download=1",
}


def download_wikigraph(wg_name: str):
    """
    Download and install a `WikiGraph`.

    Parameters
    ----------
    wg_name : str
        Name of the `WikiGraph` to download.
    """
    if wg_name not in WIKIGRAPHS_TABLE:
        msg.fail(
            f"{wg_name} not available yet. Try with: {', '.join(WIKIGRAPHS_TABLE)}",
            exits=1,
        )
    wg_tar = f"{wg_name}.tar.gz"
    run_command(f"wget -O {wg_tar} {WIKIGRAPHS_TABLE[wg_name]}")
    run_command(
        f"{sys.executable} -m pip install --no-deps --force-reinstall --no-cache {wg_tar}"
    )
    run_command(f"rm {wg_tar}")



def main():
    import sys

    import typer
    from wasabi import msg

    from . import cli

    commands = {
        "create-wikigraph": cli.create_wikigraph,
        "download-wikigraph": cli.download_wikigraph,
        "package-wikigraph": cli.package_wikigraph,
        "profile-matcher": cli.profile_matcher,
        "profile-wikigraph-load": cli.profile_wikigraph_load,
        "profile-wikigraph-exec": cli.profile_wikigraph_exec,
    }

    if len(sys.argv) == 1:
        msg.info("Available commands", ", ".join(commands), exits=1)
    command = sys.argv.pop(1)
    sys.argv[0] = "spikex %s" % command
    if command in commands:
        typer.run(commands[command])
    else:
        available = "Available: {}".format(", ".join(commands))
        msg.fail("Unknown command: {}".format(command), available, exits=1)


if __name__ == "__main__":
    main()
